/* eslint-env node */
const path = require('path')
const CopyWebpackPlugin = require('copy-webpack-plugin')
const ImageMinimizerPlugin = require('image-minimizer-webpack-plugin')
const { WebpackManifestPlugin } = require('webpack-manifest-plugin')
const MiniCssExtractPlugin = require('mini-css-extract-plugin')
const EntrypointsPlugin = require('@cactuslab/webpack-entrypoints-plugin')
const fs = require('fs')
const xml2js = require('xml2js')
const webpack = require('webpack')
const { CleanWebpackPlugin } = require('clean-webpack-plugin')
const TerserPlugin = require('terser-webpack-plugin')
const CssMinimizerPlugin = require('css-minimizer-webpack-plugin')
const exec = require('child_process').exec
const glob = require('glob')
const BundleAnalyzerPlugin = require('webpack-bundle-analyzer').BundleAnalyzerPlugin

const OUTPUT_PATH = '../../../target/frontend'

const tomatPublishingPath = findTomcatPublishingPath()

const context = path.resolve('.')

/* We use our own node_modules folder to store Webpack loader dependencies private
   to this module, assuming we're installed in a manner that supports that (e.g. pnpm)
 */
const myNodeModules = [path.resolve(__dirname, 'node_modules')]
/* Check for pnpm installed layout where @cactuslab/supermodel-frontend-webpack is _inside_ our node_modules directory */
if (path.basename(path.dirname(path.dirname(__dirname))) === 'node_modules') {
	myNodeModules.push(path.dirname(path.dirname(__dirname)))
}

module.exports = {}

module.exports.globSync = function(pattern) {
	return glob.sync(pattern)
}

module.exports.createWebpackConfig = function(env, argv, config) {
	const devServer = process.argv[2] === 'serve'
	config.usingDevServer = devServer
	
	const developmentMode = config.developmentMode !== undefined ? config.developmentMode : devServer || (argv.mode === 'development')
	config.developmentMode = developmentMode

	return {
		mode: developmentMode ? 'development' : 'production', 
		bail: true,
		context: context,
		devtool: config.useSourceMaps ? 'source-map' : false,
		entry: defaultConfig(config.entry, config, {
			'css/main.min': ['./css/main.less'],
			'js/main.min': ['./js/main'],
		}),
		resolve: {
			extensions: ['.tsx', '.ts', '.js', '.jsx'],
			...(typeof config.resolve === 'function' ? config.resolve(config) : typeof config.resolve === 'object' ? config.resolve : {}),
		},
		resolveLoader: {
			/* Resolve loader modules from this package as well as from the project
			   in which webpack is being run. This is what lets us use webpack loaders
			   as dependencies of this module, rather than of the project in which
			   webpack is being run.
			   See https://webpack.js.org/configuration/resolve/#resolveloader
			 */
			modules: ['node_modules', ...myNodeModules],
		},
		output: {
			filename: developmentMode ? '[name].js' : '[name].[contenthash:8].js',
			path: path.resolve(OUTPUT_PATH),
			assetModuleFilename: developmentMode ? '[path][name][ext]' : '[path][name].[contenthash:8][ext]',
			/* Output JavaScript chunks into their own directory */
			chunkFilename: 'js/chunk/[id].[chunkhash:8].js',
			/* We cannot use publicPath 'auto' on older browsers, e.g. IE 11, as it's not supported
			   see https://github.com/webpack/webpack/blob/master/lib/runtime/AutoPublicPathRuntimeModule.js#L58
			   so we use an appropriate publicPath for production, and correct it using __webpack_public_path__.
			 */
			publicPath: '/static/',
			...(config.output || {}),
		},
		optimization: {
			minimizer: [
				new TerserPlugin(),
				new CssMinimizerPlugin(),
				/* https://webpack.js.org/plugins/image-minimizer-webpack-plugin/ */
				new ImageMinimizerPlugin({
					/* Avoid any plugin errors from killing the build */
					severityError: 'warning',
					minimizer: {
						implementation: ImageMinimizerPlugin.imageminMinify,
						options: {
							plugins: [
								['gifsicle', { interlaced: true }],
								['jpegtran', { progressive: true }],
								['optipng', { optimizationLevel: 0 }],
								[
									'svgo',
									{
										plugins: [
											{
												name: 'removeViewBox',
												active: false,
											},
										],
									},
								],
							],
						},
					},
				})
			],
			...(config.optimization || {}),
		},
		plugins: plugins(config),
		stats: {
			all: false,
			builtAt: true,
			errors: true,
			errorDetails: true,
			timings: true,
			warnings: true,
			...(config.stats || {}),
		},
		module: {
			rules: [
				{
					test: [/\.js$/],
					exclude: /node_modules/,
					use: ['source-map-loader'],
					enforce: 'pre',
				},
				{
					oneOf: [
						{
							test: [/modernizrrc\.js/],
							exclude: /node_modules/,
							loader: 'webpack-modernizr-loader',
						},
						{
							test: [/\.tsx?$/],
							exclude: config.babel && config.babel.includeNodeModules ? new RegExp(`node_modules/(?!${config.babel.includeNodeModules.map(s => s + '/').join('|')})`) : /node_modules/,
							use: [
								{
									loader: 'babel-loader',
								},
								{
									loader: 'ts-loader',
									options: {
										// transpileOnly: true,
									},
								},
							],
						},
						{
							test: [/\.jsx?$/],
							exclude: config.babel && config.babel.includeNodeModules ? new RegExp(`node_modules/(?!${config.babel.includeNodeModules.map(s => s + '/').join('|')})`) : /node_modules/,
							use: [
								{
									loader: 'babel-loader',
								},
							],
						},
						...createCssLoader(config),
						...createLessLoader(config),
						...createSassLoader(config),
						/* Don't hash fonts */
						{
							test: [/\/fonts?\//],
							exclude: /node_modules/,
							type: 'asset/resource',
							generator: {
								filename: '[path][name][ext]',
							},
						},
						/* Copy and hash other files */
						{
							test: [/\.png$/, /\.jpe?g$/, /\.gif$/, /\.svg$/],
							type: 'asset',
							parser: {
								dataUrlCondition: { maxSize: 256 },
							},
						},
						{
							exclude: /node_modules/,
							type: 'asset',
							parser: {
								dataUrlCondition: { maxSize: 256 },
							},
						},
					],
				},
			],
		},
		devServer: {
			compress: true,
			port: 8088,
			devMiddleware: {
				publicPath: '/static/',
				stats: 'minimal',
			},
			allowedHosts: 'all',
			headers: {
				'Access-Control-Allow-Origin': '*',
			},
			hot: false,
			...(config.devServer || {}),
		},
	}
}

/**
 * If the given funcOrValue exists, resolve it with the config by either calling the func with the config
 * as an argument, or return it if it's not. If funcOrValue is falsey then do the same with the
 * defaultFuncOrValue.
 * @param funcOrValue
 * @param config
 * @param defaultFuncOrValue
 * @returns
 */
function defaultConfig(funcOrValue, config, defaultFuncOrValue) {
	return handleConfig(funcOrValue || defaultFuncOrValue, config, (result) => result)
}

/**
 * Evaluate the given funcOrValue (call the function with the config as an argument if it's a function)
 * and then pass it to the handler function, and return the result.
 * @param funcOrValue
 * @param config
 * @param handler
 * @returns
 */
function handleConfig(funcOrValue, config, handler) {
	if (typeof funcOrValue === 'function') {
		return handler(funcOrValue(config))
	} else {
		return handler(funcOrValue)
	}
}

function findCopyPluginConfig(pluginConfig, from) {
	return pluginConfig.find(row => row.from === from)
}

function plugins(config) {
	const developmentMode = config.developmentMode
	
	const result = []

	/* https://github.com/johnagan/clean-webpack-plugin */
	if (!config.usingDevServer) {
		result.push(new CleanWebpackPlugin())
	}

	const copyConfig = handleConfig(config.copy, config, (pluginConfig) => {
		if (!pluginConfig) {
			pluginConfig = []
		}
		
		if (!findCopyPluginConfig(pluginConfig, './font') && glob.sync(`${path.resolve(context, 'font')}/**`, { nodir: true }).length > 0) {
			pluginConfig.push({
				from: './font',
				to: 'font',
			})
		}
		if (!findCopyPluginConfig(pluginConfig, './fonts') && glob.sync(`${path.resolve(context, 'fonts')}/**`, { nodir: true }).length > 0) {
			pluginConfig.push({
				from: './fonts',
				to: 'fonts',
			})
		}
		if (!findCopyPluginConfig(pluginConfig, './img') && glob.sync(`${path.resolve(context, 'img')}/**`, { nodir: true }).length > 0) {
			pluginConfig.push({
				/* Copy images into our bundle as we may refer to them directly from templates */
				from: './img',
				/* We match the hash that gets applied by the assetModuleFilename
				   so that we don't duplicate images that have already come across from being
				   referenced in CSS.
				 */
				to: developmentMode ? 'img/[path][name][ext]' : 'img/[path][name].[contenthash:8][ext]',
			})
		}
		return {
			patterns: pluginConfig,
		}
	})
	if (copyConfig.patterns.length) {
		result.push(new CopyWebpackPlugin(copyConfig))
	}

	result.push(new MiniCssExtractPlugin({
		filename: developmentMode ? '[name].css' : '[name].[contenthash:8].css',
		chunkFilename: 'css/chunk/[id].[chunkhash:8].css',
		/* If we import CSS in JavaScript then a situation might arise where the order in which
		   the CSS is imported might be in question... such as if there are asynchronously loaded
		   parts of the JavaScript and modules may be imported in varying orders... we assume that
		   we design our CSS to be robust in this case by properly namespacing all of these dynamic
		   CSS imports. See https://github.com/facebook/create-react-app/issues/5372#issuecomment-685932009
		 */
		ignoreOrder: true,
	}))
	result.push(new webpack.ProvidePlugin(defaultConfig(config.provide, config, {})))

	result.push(new WebpackManifestPlugin({
		basePath: '/static/',
		publicPath: '/static/', /* Workaround for https://github.com/shellscape/webpack-manifest-plugin/pull/257 */
	}))
	result.push(new EntrypointsPlugin({
		path: 'entrypoints.json',
	}))

	result.push({
		apply: (compiler) => {
			compiler.hooks.afterEmit.tap('AfterEmitPlugin', () => {
				if (config.usingDevServer) {
					return
				}
				
				if (fs.existsSync(path.resolve(OUTPUT_PATH))) {
					if (tomatPublishingPath) {
						console.log(`Syncing files to published Tomcat webapp: ${tomatPublishingPath}`)
						exec(`rsync -a ${path.resolve(OUTPUT_PATH)}/ ${tomatPublishingPath}`)
					}
					
					exec(`rsync -a ${path.resolve(OUTPUT_PATH)}/ target/m2e-wtp/web-resources/static/`)
				}
			})
		},
	})
	result.push(new webpack.DefinePlugin(config.define || {}))

	if (!config.usingDevServer) {
		result.push(new BundleAnalyzerPlugin({
			analyzerMode: 'static',
			reportFilename: '../frontend-report.html',
			defaultSizes: 'gzip',
			openAnalyzer: false,
			...(config.bundleAnalyzer || {}),
		}))
	}

	/* Add custom plugins from the config */
	if (config.plugins) {
		if (typeof config.plugins === 'function') {
			result.push(...(config.plugins(config) || []))
		} else {
			result.push(...config.plugins)
		}
	}

	return result
}

function findWtpWebapps(searchPath) {
	if (searchPath === undefined) {
		searchPath = process.cwd()
	}

	if (fs.existsSync(searchPath + '/wtpwebapps')) {
		return searchPath + '/wtpwebapps'
	}

	var wstCorePath = searchPath + '/.metadata/.plugins/org.eclipse.wst.server.core'
	if (fs.existsSync(wstCorePath)) {
		var files = fs.readdirSync(wstCorePath).sort().reverse()
		for (var i = 0; i < files.length; i++) {
			if (files[i].indexOf('tmp') === 0) {
				if (fs.existsSync(wstCorePath + '/' + files[i] + '/wtpwebapps')) {
					return wstCorePath + '/' + files[i] + '/wtpwebapps'
				}
			}
		}
	}

	var nextSearchPath = path.dirname(searchPath)
	if (nextSearchPath !== searchPath) {
		return findWtpWebapps(nextSearchPath)
	} else {
		return false
	}
}

function findTomcatPublishingPath() {
	var wtpWebapps = findWtpWebapps()
	if (wtpWebapps === false) {
		return false
	}

	/* Use directory name as the project code in the wtpwebapps directory */
	var mavenProjectDir = findMavenProjectDir()
	var projectCode = path.basename(mavenProjectDir)
	var publishingPath = wtpWebapps + '/' + projectCode
	if (fs.existsSync(publishingPath)) {
		return publishingPath + '/static'
	}
	
	/* Use the artifactId as the project code in the wtpwebapps directory */
	var pomXml = fs.readFileSync(path.resolve(mavenProjectDir, 'pom.xml'), 'UTF-8')
	xml2js.parseString(pomXml, function(err, result) {
		projectCode = result.project.artifactId[0]
	})
	publishingPath = wtpWebapps + '/' + projectCode
	if (fs.existsSync(publishingPath)) {
		return publishingPath + '/static'
	}
	
	return false
}

function findMavenProjectDir(searchPath) {
	if (searchPath === undefined) {
		searchPath = process.cwd()
	}

	if (fs.existsSync(searchPath + '/pom.xml')) {
		return searchPath
	}

	var nextSearchPath = path.dirname(searchPath)
	if (nextSearchPath !== searchPath) {
		return findMavenProjectDir(nextSearchPath)
	} else {
		return false
	}
}

function cssLoaders(config) {
	return [
		{
			loader: MiniCssExtractPlugin.loader,
			options: {
				/* We always use publicPath 'auto' with mini-css-extract-plugin as it perfectly resolves urls relative to
				the output css file.
				*/
				publicPath: 'auto',
			},
		},
		{
			/* https://github.com/webpack-contrib/css-loader */
			loader: 'css-loader',
			options: {
				importLoaders: 1,
				import: false, /* So we don't resolve @imports in css files, as they should refer to files that will exist. */
				sourceMap: config.useSourceMaps,
				url: {
					/* https://github.com/webpack-contrib/css-loader/issues/1367 */
					filter: (url) => {
						if (/^data:/.test(url)) {
							return false;
						}
						return true;
					},
				},
			},
		},
		{
			loader: 'postcss-loader',
			options: {
				postcssOptions: {
					ident: 'postcss',
					plugins: [
						['autoprefixer', {}],
					],
				},
				sourceMap: config.useSourceMaps,
			},
		},
	]
}

function createCssLoader(config) {
	return [
		{
			test: /\.css$/,
			use: cssLoaders(config),
		},
	]
}

function createLessLoader(config) {
	/* Check if less-loader is available */
	try {
		require('less-loader')
	} catch (error) {
		return []
	}

	return [
		{
			test: /\.less$/,
			use: [
				...cssLoaders(config),
				{
					loader: 'less-loader',
					options: {
						lessOptions: {
							noIeCompat: true,
							relativeUrls: false, /* Necessary to resolve images? */
						},
						sourceMap: config.useSourceMaps,
					},
				},
			],
		},
	]
}

function createSassLoader(config) {
	/* Check if sass-loader is available */
	try {
		require('sass-loader')
	} catch (error) {
		return []
	}

	return [
		{
			test: /\.(sass)$/,
			use: [
				...cssLoaders(config),
				{
					loader: 'sass-loader',
					options: {
						sourceMap: config.useSourceMaps,
					},
				},
			],
		},
	]
}
